%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "glbtree.h"
#include "gfuncs.h"
#include "raichu_alu.h"
#include "raichu_fpu.h"

#ifdef WIN32
#define __STDC__
#endif

int yylex();
int yyparse();

int existserr = 0;
struct glblets *glets_root = NULL;
unsigned int unidentnum = 0;

extern int yylineno;

void yyerror(const char *str){
	fprintf(stderr, "line(%d) parse error: %s\n", yylineno, str);
	existserr = 1;
}

int yywrap(){
	return 1;
}


%}

%union{
	char *name;
	unsigned int val;
	struct glblets *glets;
	struct gexplist *gel;
	struct glb_exp *gexp;
}

%token <val> BOOL_VAL
%token <val> INT_VAL
%token <val> FLOAT_VAL

%token LPAREN
%token RPAREN
%token LET
%token IN
%token <name> IDENT
%token ARRAY_CREATE
%token COMMA
%token MINUS
%token EQUAL
%token DOT

%type <gexp> program 
%type <glets> lets
%type <gexp> inlets exp ary tplbd simple_exp
%type <gexplist> tplbd_


%left PREC_PARENL
%left PREC_PARENM
%left PREC_PARENH


%%

program:
	lets {
		$<glets>$ = $<glets>1;
		glets_root = $<glets>1;
	}
;

lets:
	LET IDENT EQUAL inlets lets{
		struct glblets *gl = 
			new_glblets($<name>2, $<gexp>4, $<glets>5);
		$<glets>$ = gl;
	}
|	LET IDENT EQUAL inlets {
		struct glblets *gl = 
			new_glblets($<name>2, $<gexp>4, NULL);
		$<glets>$ = gl;
	}
|	error IN {
		yyerrok;
	}
;

inlets:
	exp {
		$<gexp>$ = $<gexp>1;
	}
|	LET IDENT EQUAL exp IN inlets {
		struct glb_exp *ge_hd, *ge_tl, **ge_tlp;
		
		ge_hd = $<gexp>4;
		ge_tl = $<gexp>4;
		ge_tlp = &ge_hd;
		
		while(1){
			if((ge_tl->sort) == EXP_LET){
				ge_tlp = &((ge_tl->body).letb.next);
				ge_tl = (ge_tl->body).letb.next;
			}
			else
				break;
		}
		
		(*ge_tlp) = new_gexp(EXP_LET);
		
		((*ge_tlp)->body).letb.vname = $<name>2;
		((*ge_tlp)->body).letb.gexp = ge_tl;
		((*ge_tlp)->body).letb.next = $<gexp>6;
		
		$<gexp>$ = ge_hd;
	}
;


exp:
	tplbd {
		$<gexp>$ = $<gexp>1;
	}
|	ary {
		$<gexp>$ = $<gexp>1;
	}
;


ary:
	LPAREN ary RPAREN {
		$<gexp>$ = $<gexp>2;
	}
|	ARRAY_CREATE INT_VAL simple_exp {
		struct glb_exp *ge = new_gexp(EXP_ARRAY);
		(ge->body).arrayb.lenn = $<val>2;
		(ge->body).arrayb.initvaln = $<gexp>3;
		$<gexp>$ = ge;
	}
|	ARRAY_CREATE INT_VAL ary {
		struct glb_exp *ge_var = new_gexp(EXP_VAR);
		struct glb_exp *ge_ary = new_gexp(EXP_ARRAY);
		struct glb_exp *ge_hd, *ge_tl, **ge_tlp;
		char *newnm;
		
		newnm = (char*)malloc_noerr(sizeof(char) * 11);
		
		sprintf(newnm, "__%08x", unidentnum);
		unidentnum++;
		if(!unidentnum){
			fprintf(stderr, "create_array used too many times\n");
			exit(1);
		}
		
		ge_hd = $<gexp>3;
		ge_tl = $<gexp>3;
		ge_tlp = &ge_hd;
		
		while(1){
			if((ge_tl->sort) == EXP_LET){
				ge_tlp = &((ge_tl->body).letb.next);
				ge_tl = (ge_tl->body).letb.next;
			}
			else
				break;
		}
		
		(*ge_tlp) = new_gexp(EXP_LET);
		
		((*ge_tlp)->body).letb.vname = newnm;
		((*ge_tlp)->body).letb.gexp = ge_tl;
		((*ge_tlp)->body).letb.next = ge_ary;
		
		(ge_var->body).varb.vname = copy_string(newnm);
		
		(ge_ary->body).arrayb.lenn = $<val>2;
		(ge_ary->body).arrayb.initvaln = ge_var;
		
		$<gexp>$ = ge_hd;
	}
|	ARRAY_CREATE INT_VAL tplbd {
		struct glb_exp *ge_var = new_gexp(EXP_VAR);
		struct glb_exp *ge_ary = new_gexp(EXP_ARRAY);
		struct glb_exp *ge_let = new_gexp(EXP_LET);
		char *newnm;
		
		newnm = (char*)malloc_noerr(sizeof(char) * 11);
		
		sprintf(newnm, "__%08x", unidentnum);
		unidentnum++;
		if(!unidentnum){
			fprintf(stderr, "create_array used too many times\n");
			exit(1);
		}
		
		(ge_let->body).letb.vname = newnm;
		(ge_let->body).letb.gexp = $<gexp>3;
		(ge_let->body).letb.next = ge_ary;
		
		(ge_var->body).varb.vname = copy_string(newnm);
		
		(ge_ary->body).arrayb.lenn = $<val>2;
		(ge_ary->body).arrayb.initvaln = ge_var;
		
		$<gexp>$ = ge_let;
	}
;


simple_exp:
	IDENT DOT LPAREN INT_VAL RPAREN {
		struct glb_exp *ge = new_gexp(EXP_GET);
		(ge->body).getb.aname = $<name>1;
		(ge->body).getb.index = $<val>4;
		$<gexp>$ = ge;
	}
|	LPAREN simple_exp RPAREN {
		$<gexp>$ = $<gexp>2;
	}
|	BOOL_VAL {
		struct glb_exp *ge = new_gexp(EXP_CONST);
		(ge->body).constb.val = $<val>1;
		$<gexp>$ = ge;
	}
|	MINUS INT_VAL {
		struct glb_exp *ge = new_gexp(EXP_CONST);
		(ge->body).constb.val = alu_neg($<val>2);
		$<gexp>$ = ge;
	}
|	INT_VAL {
		struct glb_exp *ge = new_gexp(EXP_CONST);
		(ge->body).constb.val = $<val>1;
		$<gexp>$ = ge;
	}
|	MINUS FLOAT_VAL {
		struct glb_exp *ge = new_gexp(EXP_CONST);
		(ge->body).constb.val = fpu_fneg($<val>2);
		$<gexp>$ = ge;
	}
|	FLOAT_VAL {
		struct glb_exp *ge = new_gexp(EXP_CONST);
		(ge->body).constb.val = $<val>1;
		$<gexp>$ = ge;
	}
|	IDENT {
		struct glb_exp *ge = new_gexp(EXP_VAR);
		(ge->body).varb.vname = $<name>1;
		$<gexp>$ = ge;
	}
;


tplbd:
	tplbd_ {
		struct glb_exp *ge = new_gexp(EXP_TPL);
		(ge->body).tplb.comps = $<gel>1;
		$<gexp>$ = ge;
	}
|	LPAREN tplbd RPAREN {
		$<gexp>$ = $<gexp>2;
	}
;


tplbd_:
	simple_exp COMMA tplbd_ {
		$<gel>$ = new_gexplist($<gexp>1, $<gel>3);
	}
|	simple_exp COMMA simple_exp {
		$<gel>$ = new_gexplist($<gexp>1, 
			new_gexplist($<gexp>3, NULL));
	}
;


%%


int main(int argc, char *argv[]){
	extern FILE *yyin;
	char *outfilename;
	
	FILE *fp;
	
	if(sizeof(unsigned int) != 4){
		fprintf(stderr, "sizeof(unsigned int) must be 4\n");
		exit(1);
	}
	
	if(((0x80000001U >> 1) != 0x40000000U) || 
	   ((0x40000000U << 1) != 0x80000000U) || 
	   ((0x80000000U << 1) != 0x00000000U)){
		fprintf(stderr, "sll and srl needed\n");
		return 1;
	}
	
	if(((-1) >> 1) != -1){
		fprintf(stderr, "sra needed\n");
		return 1;
	}
	
	if(argc < 2){
		yyin = stdin;
		fp = NULL;
		outfilename = (char*)malloc_noerr(sizeof(char) * 6);
		strcpy(outfilename, "glb.s");
	}
	else{
		fp = fopen(argv[1], "r");
		if(fp){
			char *s;
			
			yyin = fp;
			outfilename = (char*)malloc_noerr(sizeof(char) * (strlen(argv[1]) + 3));
			strcpy(outfilename, argv[1]);
			for(s = outfilename; (*s) && (*s != '.'); s++);
			*s = '.';
			*(s + 1) = 's';
			*(s + 2) = '\0';
		}
		else{
			yyin = stdin;
			outfilename = (char*)malloc_noerr(sizeof(char) * 9);
			strcpy(outfilename, "output.s");
		}
	}
	
	yyparse();
	
	if(fp)
		fclose(fp);
	
	if(!existserr)
		compile(glets_root, outfilename);
	
	free(outfilename);
	
	return 0;
}
