#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "glbtree.h"
#include "gfuncs.h"
#include "raichu_fpu.h"


struct gexplist *new_gexplist(struct glb_exp *ge, struct gexplist *nxt){
	struct gexplist *ngel;
	
	ngel = (struct gexplist *)malloc_noerr(sizeof(struct gexplist));
	(ngel->gexp) = ge;
	(ngel->next) = nxt;
	
	return ngel;
}

void delete_gexplist(struct gexplist *gel){
	if(gel){
		delete_gexplist(gel->next);
		delete_gexp(gel->gexp);
		free(gel);
	}
}

struct glb_exp *new_gexp(enum EXP_SORT gnsort){
	struct glb_exp *nge;
	
	nge = (struct glb_exp *)malloc_noerr(sizeof(struct glb_exp));
	
	(nge->sort) = gnsort;
	
	return nge;
}

void delete_gexp(struct glb_exp *ge){
	if(ge){
		switch(ge->sort){
		case EXP_CONST:
			break;
		case EXP_VAR:
			free((ge->body).varb.vname);
			break;
		case EXP_TPL:
			delete_gexplist((ge->body).tplb.comps);
			break;
		case EXP_LET:
			delete_gexp((ge->body).letb.gexp);
			delete_gexp((ge->body).letb.next);
			free((ge->body).letb.vname);
			break;
		case EXP_ARRAY:
			delete_gexp((ge->body).arrayb.initvaln);
			break;
		case EXP_GET:
			free((ge->body).getb.aname);
			break;
		default:
			fprintf(stderr, "delete_gexp: internal error(unknown node)\n");
		}
		
		free(ge);
	}
}


struct glb_exp *copy_simplegexp(struct glb_exp *ge){
	struct glb_exp *ret;
	
	switch(ge->sort){
	case EXP_CONST:
		ret = new_gexp(EXP_CONST);
		(ret->body).constb.val = (ge->body).constb.val;
		break;
	case EXP_VAR:
		ret = new_gexp(EXP_VAR);
		(ret->body).varb.vname = copy_string((ge->body).varb.vname);
		break;
	default:
		fprintf(stderr, "copy_simplegexp: internal error(unknown node or complicated node)\n");
		ret = NULL;
	}
	
	return ret;
}


struct glblets *new_glblets(char *vn, struct glb_exp *ge, struct glblets *nxt){
	struct glblets *ngl;
	
	ngl = (struct glblets *)malloc_noerr(sizeof(struct glblets));
	
	(ngl->vname) = vn;
	(ngl->gexp) = ge;
	(ngl->next) = nxt;
	
	return ngl;
}

void delete_glblets(struct glblets *gl){
	if(gl){
		free(gl->vname);
		delete_gexp(gl->gexp);
		delete_glblets(gl->next);
		free(gl);
	}
}

void var_rename(char **strp, unsigned int id){
	int oldvnlen;
	char *newname;
	
	oldvnlen = strlen(*strp);
	newname = (char*)malloc_noerr(sizeof(char) * (oldvnlen + 10));
	strcpy(newname, *strp);
	sprintf(&(newname[oldvnlen]), "$%08x", id);
	free(*strp);
	*strp = newname;
}

int var_check(struct glb_exp *ge, char **extenvs, unsigned int extnum, 
                                  char **lenvs, unsigned int lst, unsigned int lnum){
	unsigned int i;
	char **aimstr = NULL;
	
	switch(ge->sort){
	case EXP_CONST:
		return 0;
	case EXP_VAR:
		aimstr = &((ge->body).varb.vname);
		break;
	case EXP_TPL:{
		struct gexplist *gl;
		
		for(gl = (ge->body).tplb.comps; gl; gl = (gl->next)){
			if(var_check(gl->gexp, extenvs, extnum, lenvs, lst, lnum))
				return 1;
		}
		
		return 0;
	}
	case EXP_ARRAY:
		return var_check((ge->body).arrayb.initvaln, extenvs, extnum, lenvs, lst, lnum);
	case EXP_GET:
		aimstr = &((ge->body).getb.aname);
		break;
	default:
		fprintf(stderr, "var_check: internal error(unknown node or let)\n");
		return 1;
	}
	
	for(i = lnum - 1; (i >= lst) && (i < lnum); i--){
		if(!(strcmp(lenvs[i], *aimstr)))
			break;
	}
	
	if((i >= lst) && (i < lnum))
		var_rename(aimstr, i);
	else{
		for(i = extnum - 1; i < extnum; i--){
			if(!(strcmp(extenvs[i], *aimstr)))
				break;
		}
		
		if(i >= extnum){
			fprintf(stderr, "variable %s not found\n", *aimstr);
			return 1;
		}
	}
	
	return 0;
}

struct glb_exp *search_arraybelem(char *tstr, struct glblets *gl, struct glb_exp *inletstop){
	struct glb_exp *target = NULL;
	
	if(strstr(tstr, "$")){
		for(; (inletstop->sort) == EXP_LET; inletstop = (inletstop->body).letb.next){
			if(!(strcmp(tstr, (inletstop->body).letb.vname))){
				target = (inletstop->body).letb.gexp;
				break;
			}
		}
	}
	else{
		for(; gl; gl = (gl->next)){
			if(!(strcmp(tstr, (gl->vname)))){
				for(target = gl->gexp; (target->sort) == EXP_LET; target = (target->body).letb.next);
			}
		}
	}
	
	if(!target){
		fprintf(stderr, "search_arraybelem: variable not found\n");
		return NULL;
	}
	
	if((target->sort) == EXP_ARRAY)
		return ((target->body).arrayb.initvaln);
	
	return NULL;
}

int expget_expand(struct glb_exp **gep, struct glblets *gl, struct glb_exp *inletstop){
	switch((*gep)->sort){
	case EXP_CONST:
	case EXP_VAR:
		break;
	case EXP_TPL:{
		struct gexplist *glst;
		
		for(glst = ((*gep)->body).tplb.comps; glst; glst = (glst->next))
			expget_expand(&(glst->gexp), gl, inletstop);
		
		break;
	}
	case EXP_ARRAY:
		expget_expand(&(((*gep)->body).arrayb.initvaln), gl, inletstop);
		break;
	case EXP_GET:{
		struct glb_exp *bdy;
		
		bdy = search_arraybelem(((*gep)->body).getb.aname, gl, inletstop);
		
		if(bdy){
			delete_gexp(*gep);
			*gep = copy_simplegexp(bdy);
		}
		else{
			fprintf(stderr, "%s is not array\n", ((*gep)->body).getb.aname);
			return 1;
		}
		
		break;
	}
	default:
		fprintf(stderr, "internal error\n");
		return 1;
	}
	
	return 0;
}

int glblets_expand(struct glblets *gl){
	struct glblets *glcpy = gl;
	
	for(; gl; gl = (gl->next)){
		struct glb_exp *ge, **gep;
		
		for(gep = &(gl->gexp), ge = (gl->gexp); 
		    (ge->sort) == EXP_LET; 
		    gep = &((ge->body).letb.next), ge = (ge->body).letb.next){
			if(expget_expand(&((ge->body).letb.gexp), glcpy, (gl->gexp)))
				return 1;
		}
		
		if(expget_expand(gep, glcpy, (gl->gexp)))
			return 1;
	}
	
	return 0;
}

int glblets_check(struct glblets *gl){
	struct glblets *glcpy = gl;
	unsigned int lnum = 0, lnummax = 16;
	unsigned int extnum = 0, extnummax = 16;
	char **lenvs, **extenvs;
	
	lenvs = (char **)malloc_noerr(sizeof(char *) * lnummax);
	extenvs = (char **)malloc_noerr(sizeof(char *) * extnummax);
	
	for(; gl; gl = (gl->next)){
		unsigned int ltind = lnum;
		unsigned int i;
		struct glb_exp *ge;
		
		for(ge = (gl->gexp); (ge->sort) == EXP_LET; ge = (ge->body).letb.next){
			if(var_check((ge->body).letb.gexp, extenvs, extnum, lenvs, ltind, lnum)){
				free(extenvs);
				for(lnum--; lnum < lnummax; lnum--)
					free(lenvs[lnum]);
				free(lenvs);
				return 1;
			}
			
			if(lnum == lnummax){
				lnummax = lnummax * 2;
				lenvs = (char **)realloc_noerr(lenvs, sizeof(char *) * lnummax);
			}
			
			lenvs[lnum] = copy_string((ge->body).letb.vname);
			var_rename(&((ge->body).letb.vname), lnum);
			lnum++;
		}
		
		if(var_check(ge, extenvs, extnum, lenvs, ltind, lnum)){
			free(extenvs);
			for(lnum--; lnum < lnummax; lnum--)
				free(lenvs[lnum]);
			free(lenvs);
			return 1;
		}
		
		for(i = 0; i < extnum; i++){
			if(!(strcmp(extenvs[i], (gl->vname))))
				break;
		}
		
		if(i < extnum){
			fprintf(stderr, "more than one definition of %s exists\n", extenvs[i]);
			free(extenvs);
			for(lnum--; lnum < lnummax; lnum--)
				free(lenvs[lnum]);
			free(lenvs);
			return 1;
		}
		
		if(extnum == extnummax){
			extnummax = extnummax * 2;
			extenvs = (char **)realloc_noerr(extenvs, sizeof(char *) * extnummax);
		}
		
		extenvs[extnum] = gl->vname;
		extnum++;
	}
	
	free(extenvs);
	for(lnum--; lnum < lnummax; lnum--)
		free(lenvs[lnum]);
	free(lenvs);
	
	return glblets_expand(glcpy);
}

void fprintf_gexp(FILE *fp, struct glb_exp *ge){
	switch(ge->sort){
	case EXP_CONST:
		fprintf(fp, ".word %d;\n", (ge->body).constb.val);
		break;
	case EXP_VAR:
		fprintf(fp, ".word %s;\n", (ge->body).varb.vname);
		break;
	case EXP_TPL:{
		struct gexplist *gl;
		
		for(gl = (ge->body).tplb.comps; gl; gl = (gl->next))
			fprintf_gexp(fp, gl->gexp);
		
		break;
	}
	case EXP_ARRAY:{
		unsigned int i;
		
		for(i = 0; i < (ge->body).arrayb.lenn; i++)
			fprintf_gexp(fp, (ge->body).arrayb.initvaln);
		
		break;
	}
	default:
		fprintf(stderr, "internal error(fprintf_gexp: unknown node or let or get)\n");
	}
}

void fprintf_glblets(FILE *fp, struct glblets *gl){
	fprintf(fp, ":$DATA_BEGIN$;\n");
	
	for(; gl; gl = (gl->next)){
		struct glb_exp *ge;
		
		for(ge = (gl->gexp); (ge->sort) == EXP_LET; ge = (ge->body).letb.next){
			fprintf(fp, ":%s;\n", (ge->body).letb.vname);
			fprintf_gexp(fp, (ge->body).letb.gexp);
		}
		
		fprintf(fp, ":%s;\n", (gl->vname));
		fprintf_gexp(fp, ge);
	}
	
	fprintf(fp, ":$DATA_END$;\n");
}

int compile(struct glblets *gl, char *outfilename){
	FILE *fp;
	
	if((fp = fopen(outfilename, "w"))){
		if(!glblets_check(gl)){
			fprintf_glblets(fp, gl);
		}
		
		fclose(fp);
	}
	else
		fprintf(stderr, "cannot create file %s\n", outfilename);
	
	delete_glblets(gl);
	
	return 1;
}
